@@ -41,7 +41,9 @@ const LOG_LEVEL_ENV: &str = "RIO_LOG_LEVEL";
 
 pub fn setup_environment_variables(config: &rio_backend::config::Config) {
     #[cfg(unix)]
-    let terminfo = if teletypewriter::terminfo_exists("rio") {
+    let terminfo = if teletypewriter::terminfo_exists("rio")
+        || teletypewriter::terminfo_exists("xterm-rio")
+    {
         "rio"
     } else {
         "xterm-256color"
@@ -179,10 +181,32 @@ fn main() -> Result<(), Box<dyn std::error::Error>> {
         }
 
         if let Some(working_dir_cli) = args.window_options.terminal_options.working_dir {
-            config.working_dir = Some(working_dir_cli);
+            config.working_dir = match std::fs::canonicalize(&working_dir_cli).and_then(
+                |path| {
+                    if path.is_dir() {
+                        path.into_os_string().into_string().map_err(|_| {
+                            std::io::Error::new(
+                                std::io::ErrorKind::InvalidData,
+                                "Invalid UTF-8 in path",
+                            )
+                        })
+                    } else {
+                        Err(std::io::Error::new(
+                            std::io::ErrorKind::NotADirectory,
+                            "Path is not a directory",
+                        ))
+                    }
+                },
+            ) {
+                Ok(canonical_path) => Some(canonical_path),
+                Err(e) => {
+                    tracing::warn!("Failed to set working directory '{}': {}. Using default instead.", working_dir_cli, e);
+                    None
+                }
+            };
         }
 
-        config.window.initial_title = args.window_options.terminal_options.title;
+        config.title.placeholder = args.window_options.terminal_options.title_placeholder;
     }
 
     #[cfg(target_os = "linux")]
